/*
 * Author: Dušan Kolář
 */

#ifndef PARSE_DLMS
#define PARSE_DLMS

#include "parse_hdlc.h"

typedef enum e_dlms_parse_error {
    DLMS_OK,
    DLMS_TOO_SHORT,                 // passed buffer is shorter then smallest packet
    DLMS_OUT_OF_DATA,               // running out of the passed data length
    DLMS_WRONG_PAYLOAD_MARKER,      // payload marker is either 0xfN or 0xNf, where N<0xf
    DLMS_EXPECTED_VERSION_01,       // version not 01
    DLMS_TOKEN_LENGTH_SIZE,         // token length >8
    DLMS_ZERO_LENGTH_PAYLOAD,       // payload marker present, but zero length payload present
    DLMS_OUT_OF_MEMORY,             // malloc failed to return free memory thunk
    DLMS_OPTION_OUT_OF_BOUNDS,      // option length is out of allowed limits
    DLMS_UINT_GT64,                 // uint option is larger than 64 bits
} dlms_parse_error_t;


// DLMS PDU types
#define DLMS_PDU_GET_REQUEST            192
#define DLMS_PDU_SET_REQUEST            193
#define DLMS_PDU_NOTIFICATION_REQUEST   194
#define DLMS_PDU_ACTION_REQUEST         195
#define DLMS_PDU_GET_RESPONSE           196
#define DLMS_PDU_SET_RESPONSE           197
#define DLMS_PDU_ACTION_RESPONSE        199

// DLMS request type for DLMS get request PDU
#define DLMS_GET_REQ_NORMAL             1
#define DLMS_GET_REQ_NEXT               2
#define DLMS_GET_REQ_WITH_LIST          3

// DLMS request type for DLMS set request PDU
#define DLMS_SET_REQ_NORMAL             1
#define DLMS_SET_REQ_WITH_1_DBLCK       2
#define DLMS_SET_REQ_WITH_DBLCK         3
#define DLMS_SET_REQ_WITH_LIST          4
#define DLMS_SET_REQ_W_LIST_1_DBLCK     5

// DLMS action request type
#define DLMS_ACT_REQ_NORMAL             1
#define DLMS_ACT_REQ_NEXT_PBLOCK        2
#define DLMS_ACT_REQ_LIST               3
#define DLMS_ACT_REQ_1PBLOCK            4
#define DLMS_ACT_REQ_LIST_1PBLOCK       5
#define DLMS_ACT_REQ_PBLOCK             6

// DLMS type for get response type
#define DLMS_GET_RESP_NORMAL            1
#define DLMS_GET_RESP_NEXT              2
#define DLMS_GET_RESP_LIST              3

// DLMS access result types
#define DLMS_ACC_RES_SUCCESS            0
#define DLMS_ACC_RES_HW_FAULT           1
#define DLMS_ACC_RES_TEMPORARY_FAIL     2
#define DLMS_ACC_RES_RW_DENIED          3
#define DLMS_ACC_RES_OBJ_UNDEFINED      4
#define DLMS_ACC_RES_OBJ_CLS_INCONSIST  9
#define DLMS_ACC_RES_OBJ_UNAVAIL        11
#define DLMS_ACC_RES_TYPE_UNMATCHED     12
#define DLMS_ACC_RES_ACC_SCOPE_VIOL     13
#define DLMS_ACC_RES_DAT_BLCK_UNAVAIL   14
#define DLMS_ACC_RES_LONG_GET_ABORTED   15
#define DLMS_ACC_RES_NO_LONG_GET_IN     16
#define DLMS_ACC_RES_LONG_SET_ABORTED   17
#define DLMS_ACC_RES_NO_LONG_SET_IN     18
#define DLMS_ACC_RES_OTHER_REASON       250

// DLMS set response type
#define DLMS_SET_RESP_NORMAL            1
#define DLMS_SET_RESP_DBLCK             2
#define DLMS_SET_RESP_LAST_DBLCK        3
#define DLMS_SET_RESP_LAST_DBLCK_LIST   4
#define DLMS_SET_RESP_LIST              5

// DLMS action response type
#define DLMS_ACT_RESP_NORMAL            1
#define DLMS_ACT_RESP_PBLOCK            2
#define DLMS_ACT_RESP_LIST              3
#define DLMS_ACT_RESP_NEXT_PBLOCK       4

// DLMS action result type for action response type
#define DLMS_ACT_RESP_RES_SUCCESS            0
#define DLMS_ACT_RESP_RES_HW_FAULT           1
#define DLMS_ACT_RESP_RES_TEMPORARY_FAIL     2
#define DLMS_ACT_RESP_RES_RW_DENIED          3
#define DLMS_ACT_RESP_RES_OBJ_UNDEFINED      4
#define DLMS_ACT_RESP_RES_OBJ_CLS_INCONSIST  5
#define DLMS_ACT_RESP_RES_OBJ_UNAVAIL        11
#define DLMS_ACT_RESP_RES_TYPE_UNMATCHED     12
#define DLMS_ACT_RESP_RES_ACC_SCOPE_VIOL     13
#define DLMS_ACT_RESP_RES_DAT_BLCK_UNAVAIL   14
#define DLMS_ACT_RESP_RES_LONG_ACT_ABORTED   15
#define DLMS_ACT_RESP_RES_NO_LONG_ACT_IN     16
#define DLMS_ACT_RESP_RES_OTHER_REASON       250

#include <inttypes.h>

dlms_pdu_t *parse_dlms(uint8_t **input_ptr, uint32_t size, hdlc_parse_error_t *error);

#endif
