/*
 * Author: Dušan Kolář
 */

#include <stdio.h>
#include <malloc.h>

#include "parse_dlms.h"
#include "dlms_data.h"


void print_data_access_result(uint8_t type) {
    printf("DLMS data access result: %u - ",(unsigned)type);
    switch (type) {
        case
            DLMS_ACC_RES_SUCCESS:
            puts("DLMS_ACC_RES_SUCCESS");
            break;
        case
            DLMS_ACC_RES_HW_FAULT:
            puts("DLMS_ACC_RES_HW_FAULT");
            break;
        case
            DLMS_ACC_RES_TEMPORARY_FAIL:
            puts("DLMS_ACC_RES_TEMPORARY_FAIL");
            break;
        case
            DLMS_ACC_RES_RW_DENIED:
            puts("DLMS_ACC_RES_RW_DENIED");
            break;
        case
            DLMS_ACC_RES_OBJ_UNDEFINED:
            puts("DLMS_ACC_RES_OBJ_UNDEFINED");
            break;
        case
            DLMS_ACC_RES_OBJ_CLS_INCONSIST:
            puts("DLMS_ACC_RES_OBJ_CLS_INCONSIST");
            break;
        case
            DLMS_ACC_RES_OBJ_UNAVAIL:
            puts("DLMS_ACC_RES_OBJ_UNAVAIL");
            break;
        case
            DLMS_ACC_RES_TYPE_UNMATCHED:
            puts("DLMS_ACC_RES_TYPE_UNMATCHED");
            break;
        case
            DLMS_ACC_RES_ACC_SCOPE_VIOL:
            puts("DLMS_ACC_RES_ACC_SCOPE_VIOL");
            break;
        case
            DLMS_ACC_RES_DAT_BLCK_UNAVAIL:
            puts("DLMS_ACC_RES_DAT_BLCK_UNAVAIL");
            break;
        case
            DLMS_ACC_RES_LONG_GET_ABORTED:
            puts("DLMS_ACC_RES_LONG_GET_ABORTED");
            break;
        case
            DLMS_ACC_RES_NO_LONG_GET_IN:
            puts("DLMS_ACC_RES_NO_LONG_GET_IN");
            break;
        case
            DLMS_ACC_RES_LONG_SET_ABORTED:
            puts("DLMS_ACC_RES_LONG_SET_ABORTED");
            break;
        case
            DLMS_ACC_RES_NO_LONG_SET_IN:
            puts("DLMS_ACC_RES_NO_LONG_SET_IN");
            break;
        case
            DLMS_ACC_RES_OTHER_REASON:
            puts("DLMS_ACC_RES_OTHER_REASON");
            break;
        default:
            puts("Unknown !!!");
            break;
    }
}

void print_dlms_act_resp(uint8_t type) {
    printf("DLMS action response: %u - ",(unsigned)type);
    switch (type) {
        case
            DLMS_ACT_RESP_RES_SUCCESS:
            puts("DLMS_ACT_RESP_RES_SUCCESS");
            break;
        case
            DLMS_ACT_RESP_RES_HW_FAULT:
            puts("DLMS_ACT_RESP_RES_HW_FAULT");
            break;
        case
            DLMS_ACT_RESP_RES_TEMPORARY_FAIL:
            puts("DLMS_ACT_RESP_RES_TEMPORARY_FAIL");
            break;
        case
            DLMS_ACT_RESP_RES_RW_DENIED:
            puts("DLMS_ACT_RESP_RES_RW_DENIED");
            break;
        case
            DLMS_ACT_RESP_RES_OBJ_UNDEFINED:
            puts("DLMS_ACT_RESP_RES_OBJ_UNDEFINED");
            break;
        case
            DLMS_ACT_RESP_RES_OBJ_CLS_INCONSIST:
            puts("DLMS_ACT_RESP_RES_OBJ_CLS_INCONSIST");
            break;
        case
            DLMS_ACT_RESP_RES_OBJ_UNAVAIL: 
            puts("DLMS_ACT_RESP_RES_OBJ_UNAVAIL");
            break;
        case
            DLMS_ACT_RESP_RES_TYPE_UNMATCHED: 
            puts("DLMS_ACT_RESP_RES_TYPE_UNMATCHED");
            break;
        case
            DLMS_ACT_RESP_RES_ACC_SCOPE_VIOL:
            puts("DLMS_ACT_RESP_RES_ACC_SCOPE_VIOL");
            break;
        case
            DLMS_ACT_RESP_RES_DAT_BLCK_UNAVAIL:
            puts("DLMS_ACT_RESP_RES_DAT_BLCK_UNAVAIL");
            break;
        case
            DLMS_ACT_RESP_RES_LONG_ACT_ABORTED:
            puts("DLMS_ACT_RESP_RES_LONG_ACT_ABORTED");
            break;
        case
            DLMS_ACT_RESP_RES_NO_LONG_ACT_IN:
            puts("DLMS_ACT_RESP_RES_NO_LONG_ACT_IN");
            break;
        case
            DLMS_ACT_RESP_RES_OTHER_REASON:
            puts("DLMS_ACT_RESP_RES_OTHER_REASON");
            break;
        default:
            puts("Unknown !!!");
            break;
    }
}

// .................................................................
// .................................................................


void print_id_priority(dlms_pdu_t *dlms) {
    printf("Priority: %u\n",(unsigned)dlms->priority);
    printf("Service class: %u\n",(unsigned)dlms->service_class);
    printf("Invoke ID: %u\n",(unsigned)dlms->invoke_id);
}


void print_cosem_object_instance_id(dlms_pdu_t *dlms) {
    printf("Class ID: %u\n",(unsigned)dlms->req.class_id);
    printf("COSEM object instance id: ");
    for (unsigned i=0;i<6;++i) {
        printf("%u ",(unsigned)(dlms->req.instance_id[i]));
    }
    puts("");
    printf("Attribute/method ID: %u\n",(unsigned)dlms->req.attribute_method_id);
}


void print_dlms_data_type(dlms_pdu_t *dlms) {
    if (dlms->req.data_result_type == 0) {
        printf("DLMS data type: %u - ",(unsigned)dlms->req.data_type);
        puts(text_dlms_data_type(dlms->req.data_type));
        if (dlms->req.data_type == DLMS_OCTET_STRING) {
            printf("Octet string length: %u ~ ",(unsigned)dlms->req.octet_string_len);
            unsigned max = (unsigned)dlms->req.octet_string_len < 16 ? (unsigned)dlms->req.octet_string_len : 16;
            for(unsigned u=0; u<max; ++u) printf("%02x ",dlms->req.octet_string_start[u]);
            puts("");
        }
    } else {
        printf("Access selector: %u\n",(unsigned)dlms->req.access_selector);
    }
}


void print_dlms_data_type_opt(dlms_pdu_t *dlms) {
    if (dlms->req.data_present == 0) {
        puts("No data!");
        return;
    }
    printf("Access selector: %u\n",(unsigned)dlms->req.access_selector);
    print_dlms_data_type(dlms);
}


void print_date_time_opt(dlms_pdu_t *dlms) {
    if (dlms->date_time_present == 0) {
        puts("No date, no time!");
        return;
    }
    printf("Data time data sequence: ");
    for (unsigned i=0;i<12;++i) {
        printf("%u ",(unsigned)(dlms->date_time[i]));
    }
    puts("");
}

void print_block_data_access_result(dlms_pdu_t *dlms) {
    printf("Last block: %u\n",(unsigned)dlms->last_block);
    printf("Block number: %lu\n",(long unsigned)dlms->block_number);
    switch (dlms->result_choice) {
        case 0:
            printf("String of length %u: ",dlms->str_length);
            for (unsigned i=0;i<dlms->str_length;++i) {
                putchar(dlms->str[i]);
            }
            puts("");
            break;            
        case 1:
            print_data_access_result( dlms->data_access_result );
            break;
    }
}

void print_data_access_result_list(dlms_pdu_t *dlms) {
    printf("Result list length: %u\n",(unsigned)dlms->list_length);
    for (unsigned i=0; i<dlms->list_length; ++i) {
        print_data_access_result( dlms->list_data_acces_result[i] );
    }
}


// .................................................................
// .................................................................


void print_get_request(dlms_pdu_t *dlms) {
   switch (dlms->pdu_sub_type) {
        case
            DLMS_GET_REQ_NORMAL:
            
            puts(" - Get request normal");

            print_id_priority(dlms);
            print_cosem_object_instance_id(dlms);
            print_dlms_data_type_opt(dlms);
            
            break;
            
        case
            DLMS_GET_REQ_NEXT:
            
            puts(" - Get request next");

            print_id_priority(dlms);
            printf("Block number: %lu\n",(long unsigned)dlms->block_number);
            
            break;
            
        case
            DLMS_GET_REQ_WITH_LIST:
            
            puts(" - Get request with list");

            print_id_priority(dlms);
            print_cosem_object_instance_id(dlms);
            print_dlms_data_type_opt(dlms);
                        
            break;
   } // switch
}

void print_set_request(dlms_pdu_t *dlms) {
   switch (dlms->pdu_sub_type) {
        case
            DLMS_SET_REQ_NORMAL:

            puts(" - Set request normal");
            
            print_id_priority(dlms);
            print_cosem_object_instance_id(dlms);
            
            if (dlms->req.data_present) {
                puts("Optional data: ");
                print_dlms_data_type_opt(dlms);
                puts("No hard data parsed :-(");
            } else

            if (dlms->req.data_hard_present) {
                puts("No optional data!");
                puts("Hard data: ");
                print_dlms_data_type(dlms);
            } else {
                
                puts("No optional, no hard data - strange!!!");
            }
            
            break;
            
        case
            DLMS_SET_REQ_WITH_1_DBLCK:
            
            puts(" - Set request with 1st data block");
            
            print_id_priority(dlms);
            printf("First (\?\?) byte of data block: %u\n", (unsigned)*(dlms->data_block));
            
            break;

        case
            DLMS_SET_REQ_WITH_DBLCK:

            puts(" - Set request with data block");
            
            print_id_priority(dlms);
            print_cosem_object_instance_id(dlms);

            if (dlms->req.data_present) {
                puts("Optional data: ");
                print_dlms_data_type_opt(dlms);
                puts("No data block pointer :-(");
            } else {
                puts("No optional data");
                printf("First (\?\?) byte of data block: %u\n", (unsigned)*(dlms->data_block));
            }
            
            break;
            
        case
            DLMS_SET_REQ_WITH_LIST:

            puts(" - Set request with list");
            
            print_id_priority(dlms);
            printf("Number of dlms data elements: %u\nNo data :-(\n",(unsigned)dlms->item_count);
            
            break;
            
        case
            DLMS_SET_REQ_W_LIST_1_DBLCK:
            
            puts(" - Set request with list and first data block");
            
            print_id_priority(dlms);            
            printf("Number of dlms data elements: %u\nNo data :-(\n",(unsigned)dlms->item_count);
                        
            break;
   } // switch
}

void print_notification_request(dlms_pdu_t *dlms) {
    print_date_time_opt(dlms);    
    print_cosem_object_instance_id(dlms);
    if (dlms->req.data_hard_present) {
        puts("Hard data: ");
        print_dlms_data_type(dlms);
    } else {
        puts("Strange, there should be hard data!!!");
    }    
}

void print_action_request(dlms_pdu_t *dlms) {
   switch (dlms->pdu_sub_type) {
        case
            DLMS_ACT_REQ_NORMAL:

            puts(" - Action request normal");

            print_id_priority(dlms);
            print_cosem_object_instance_id(dlms);
            print_dlms_data_type_opt(dlms);
            
            break;
        
        case
            DLMS_ACT_REQ_NEXT_PBLOCK:

            puts(" - Action request next P block");

            print_id_priority(dlms);
            printf("Block number: %lu\n",(long unsigned)dlms->block_number);            
            
            break;
        
        case
            DLMS_ACT_REQ_LIST:
            puts(" - Action request with list");
            puts("!!! No data analysis done for this sub-types of PDU !!!");
            break;
            
        case
            DLMS_ACT_REQ_1PBLOCK:
            puts(" - Action request with 1st P block");
            puts("!!! No data analysis done for this sub-types of PDU !!!");
            break;
            
        case
            DLMS_ACT_REQ_LIST_1PBLOCK:
            puts(" - Action request with list and 1st P block");
            puts("!!! No data analysis done for this sub-types of PDU !!!");
            break;
            
        case
            DLMS_ACT_REQ_PBLOCK:
            puts(" - Action request with P block");
            puts("!!! No data analysis done for this sub-types of PDU !!!");
            break;
   } // switch
}

void print_get_response(dlms_pdu_t *dlms) {
   switch (dlms->pdu_sub_type) {
        case
            DLMS_GET_RESP_NORMAL:
            
            puts(" - Get response normal");

            print_id_priority(dlms);
            
            if (dlms->req.data_hard_present) {
                puts("Hard data: ");
                print_dlms_data_type(dlms);
            } else {
                puts("Strange, there should be hard data!!!");
            }    

            break;
            
        case
            DLMS_GET_RESP_NEXT:

            puts(" - Get response next");
            
            print_id_priority(dlms);
            print_block_data_access_result(dlms);
            
            break;
            
        case
            DLMS_GET_RESP_LIST:

            puts(" - Get response with list");
            
            print_id_priority(dlms);
            print_data_access_result_list(dlms);
            
            break;
   } // switch
}

void print_set_response(dlms_pdu_t *dlms) {
   switch (dlms->pdu_sub_type) {
        case
            DLMS_SET_RESP_NORMAL:
            
            puts(" - Set response normal");
            
            print_id_priority(dlms);
            print_data_access_result( dlms->data_access_result );
            
            break;
           
        case
            DLMS_SET_RESP_DBLCK:
            
            puts(" - Set response with data block");
            
            print_id_priority(dlms);
            printf("Block number: %lu\n",(long unsigned)dlms->block_number);

            break;
           
        case
            DLMS_SET_RESP_LAST_DBLCK:
            
            puts(" - Set response with last data block");
            
            print_id_priority(dlms);
            print_data_access_result( dlms->data_access_result );
            printf("Block number: %lu\n",(long unsigned)dlms->block_number);
            
            break;
            
        case
            DLMS_SET_RESP_LAST_DBLCK_LIST:
            
            puts(" - Set response with last data block and list");
            
            print_id_priority(dlms);
            print_data_access_result_list(dlms);
            printf("Block number: %lu\n",(long unsigned)dlms->block_number);
            
            break;
            
        case
            DLMS_SET_RESP_LIST:
            
            puts(" - Set response with list");
            
            print_id_priority(dlms);
            print_data_access_result_list(dlms);
            
            break;
   } // switch
}

void print_action_response(dlms_pdu_t *dlms) {
   switch (dlms->pdu_sub_type) {
        case
            DLMS_ACT_RESP_NORMAL:
            
            puts(" - Action response normal");
            
            print_id_priority(dlms);
            print_dlms_act_resp(dlms->action_response_result);
            
            if (dlms->req.data_present) {
                puts("Optional data: ");
                print_dlms_data_type_opt(dlms);
                puts("No data access result :-(");
            } else {
                puts("No optional data");
                print_data_access_result( dlms->data_access_result );
            }

            break;
            
        case
            DLMS_ACT_RESP_PBLOCK:
            puts(" - Action response with P block");
            puts("!!! No data analysis done for this sub-types of PDU !!!");
            break;
            
        case
            DLMS_ACT_RESP_LIST:
            puts(" - Action response with list");
            puts("!!! No data analysis done for this sub-types of PDU !!!");
            break;
            
        case
            DLMS_ACT_RESP_NEXT_PBLOCK:
            puts(" - Action response next P block");
            puts("!!! No data analysis done for this sub-types of PDU !!!");
            break;
   } // switch
}


// .................................................................
// .................................................................


void demo_parse_packet(uint8_t *data, long packet_size) {
    puts("Parsing...");
    
    hdlc_parse_error_t error;
    dlms_pdu_t *dlms = parse_hdlc(data, (uint32_t)packet_size, &error);
    
    puts("Parsing completed...");
    
    if (dlms == NULL) {
        puts("Packet parsing failed");
        switch (error) {
            case HDLC_TOO_SHORT: {
                puts("Passed buffer is shorter then smallest packet of given type");
                break;
            }
            case HDLC_WRONG_MAGIC: {
                puts("Wrong magic byte 0x7E on the beginning/end");
                break;
            }
            case HDLC_WRONG_MISSING_LLC: {
                puts("Expected LLC header not found");
                break;
            }
            case HDLC_BASE_SEQ: {
                puts("BASE128 sequence representing number too long");
                break;
            }
            case HDLC_OUT_OF_DATA: {
                puts("Running out of the passed data length");
                break;
            }
            case HDLC_NOT_MATCHING_SIZE: {
                puts("Size of the pasted data and packet length do not match");
                break;
            }
            case HDLC_NOT_AN_IFRAME: {
                puts("Only I-frames carry some information/payload that is further processed");
                break;
            }
            case HDLC_DLMS_TOO_SHORT: {
                puts("DLMS pdu is too short");
                break;
            }
            case HDLC_DLMS_WRONG_PDU_TYPE: {
                puts("DLMS PDU type not recognized");
                break;
            }
            case HDLC_DLMS_WRONG_GET_REQ_TYPE: {
                puts("DLMS PDU get request type not recognized");
                break;
            }
            case HDLC_DLMS_WRONG_DATA_TYPE: {
                puts("DLMS data type unknown");
                break;
            }
            case HDLC_DLMS_WRONG_SET_REQ_TYPE: {
                puts("DLMS PDU set request type not recognized");
                break;
            }
            case HDLC_DLMS_WRONG_ACT_REQ_TYPE: {
                puts("DLMS PDU action request type not recognized");
                break;
            }
            case HDLC_DLMS_WRONG_GET_RESP_TYPE: {
                puts("DLMS PDU get response type not recognized");
                break;
            }
            case HDLC_DLMS_WRONG_ACC_RES_CHOICE: {
                puts("DLMS get response with data block access response choice not recognized");
                break;
            }
            case HDLC_DLMS_WRONG_SET_RESP_TYPE: {
                puts("DLMS PDU set response type not recognized");
                break;
            }
            case HDLC_DLMS_WRONG_ACT_RESP_TYPE: {
                puts("DLMS PDU action response type not recognized");
                break;
            }
            case HDLC_DLMS_UNSUPPORTED_LEN: {
                puts("DLMS length encoded integer of unsupported size (just 1 and 2 bytes supported)");
                break;
            }
            case HDLC_DLMS_SIZE_MISMATCH: {
                puts("Expected and derived length of packet do not match");
                break;
            }
            case HDLC_OUT_OF_MEMORY: {
                puts("Out of predefined size for variable structures");
                break;
            }
            default: {
                puts("Unknown reason");
                break;
            }
        }
        puts("");
    } else {
        switch (dlms->pdu_type) {
            case
                DLMS_PDU_GET_REQUEST:
                puts("Get request:");
                print_get_request(dlms);
                break;
            
            case
                DLMS_PDU_SET_REQUEST:
                puts("Set request:");
                print_set_request(dlms);
                break;
            
            case
                DLMS_PDU_NOTIFICATION_REQUEST:
                puts("Event notification request:");
                print_notification_request(dlms);
                break;
            
            case
                DLMS_PDU_ACTION_REQUEST:
                puts("Action request:");
                print_action_request(dlms);
                break;
            
            case
                DLMS_PDU_GET_RESPONSE:
                puts("Get response:");
                print_get_response(dlms);
                break;

            case
                DLMS_PDU_SET_RESPONSE:
                puts("Set response:");
                print_set_response(dlms);
                break;

            case
                DLMS_PDU_ACTION_RESPONSE:
                puts("Action response:");
                print_action_response(dlms);
                break;
        }
         puts("");
    }
    
}


// .................................................................
// .................................................................


int main(int argc, char *argv[]) {
    if (argc<2) {
        puts("Usage: test <hdlc+dlms_packet_file(s)>");
        return 1;
    }
    
    for (int i=1; i<argc;++i) {
        printf("Started %s...\n",argv[i]);
    
        FILE *packet;
        if ((packet = fopen(argv[i],"rb")) == NULL) {
            printf("Cannot open file: %s\n", argv[i]);
            return 2;
        }
        
        if (fseek(packet, 0L, SEEK_END)) {
            printf("Cannot work with file: %s\n",argv[i]);
            fclose(packet);
            return 3;
        }
        
        long packet_size = ftell(packet);
        
        if (packet_size<0) {
            printf("Cannot work with file: %s\n",argv[i]);
            fclose(packet);
            return 4;
        }
        
        if (packet_size > 0x10000L) {
            printf("Packet size too large: %ld\n",packet_size);
            fclose(packet);
            return 5;
        }
        
        if (fseek(packet, 0L, SEEK_SET)) {
            printf("Cannot work with file: %s\n",argv[i]);
            fclose(packet);
            return 6;
        }
        
        uint8_t *data = malloc((size_t)packet_size);
        
        if (fread((void*)data,1,(size_t)packet_size,packet) != (size_t)packet_size) {
            printf("Cannot read file: %s\n",argv[i]);
            fclose(packet);
            return 7;
        }
        
        fclose(packet);
        
        demo_parse_packet(data, packet_size);
        
        free((void*)data);
        
        printf("Done %s !\n",argv[i]);
        
    } // end for
    
    return 0;
} // main

// EOF
